'use strict';

const RecordingAPI = function ()
{

  /**
   * @param {string} query
   * @return {Object}
   */
  const objectFromQueryString = function (query)
  {
    const entries = query.split('&')
      .filter(x => x !== '')
      .map(x => x.split('=', 2));
    return Object.fromEntries(entries);
  };

  /**
   * @return {Promise<[{profile: string, diskid: string, options: Object, eventid: string}]>}
   */
  const ruleList = async function ()
  {
    const url = '/axis-cgi/record/continuous/listconfiguration.cgi';
    const result_node = await _callXmlApi(url, 'continuousrecordingconfigurations');

    const conf_nodes = result_node.getElementsByTagName('continuousrecordingconfiguration');
    const configs = [];
    for (const node of conf_nodes) {
      configs.push({
        profile: node.getAttribute('profile'),
        diskid: node.getAttribute('diskid'),
        options: objectFromQueryString(node.getAttribute('options')),
        eventid: node.getAttribute('eventid'),
      });
    }
    return configs;
  };

  /**
   *
   * @param {string} disk_id
   * @param {Object} options
   * @return {Promise<string>} profile_id
   */
  const ruleAdd = async function (disk_id, options = {}) {
    const options_string = Object.entries(options).map(x => x.join('=')).join('&');
    const params = {
      diskid: disk_id,
      options: options_string,
      eventid: 'continuous',
    };

    const url = '/axis-cgi/record/continuous/addconfiguration.cgi' + '?' + new URLSearchParams(Object.entries(params));
    const result_node = await _callXmlApi(url, 'configure');

    const result = result_node.getAttribute('result');
    if (result !== 'OK') {
      throw new Error(result_node.getAttribute('errormsg') || result);
    }

    return result_node.getAttribute('profile');
  };

  /**
   * @param {String} profile
   * @return {Promise<void>}
   */
  const ruleRemove = async function (profile)
  {
    const url = '/axis-cgi/record/continuous/removeconfiguration.cgi?profile=' + profile;
    const result_node = await _callXmlApi(url, 'remove');

    const result = result_node.getAttribute('result');
    if (result !== 'OK') {
      throw new Error(result_node.getAttribute('errormsg') || result);
    }
  };

  /**
   * Check if any of matched Edge storages is ready
   * @param {string} diskid
   * @return {Promise<boolean>}
   */
  const checkEdgeReady = async function (diskid = 'all')
  {
    const url = '/axis-cgi/disks/list.cgi?diskid=' + diskid;
    const result_node = await _callXmlApi(url, 'disks');
    const disks = result_node.getElementsByTagName('disk');
    if (disks.length === 0) {
      return false;
    }
    const properties_required_to_ready = {
      status: 'OK',
      locked: 'no',
      readonly: 'no',
    };
    for (const disk of disks) {
      let is_ready = true;
      for (const name in properties_required_to_ready) {
        const value = disk.getAttribute(name);
        is_ready = is_ready && properties_required_to_ready[name] === value;
      }
      if (is_ready) {
        return true;
      }
    }
    return false;
  };

  /**
   * @param {string} url
   * @param {string} result_node_name
   * @return {Promise<Element>}
   * @private
   */
  const _callXmlApi = async function (url, result_node_name)
  {
    const response = await fetch(url);

    if (!response.ok) {
      throw new Error(response.statusText || 'Network error');
    }

    const xml_reply = await response.text();
    const doc = new DOMParser().parseFromString(xml_reply, 'text/xml');

    const result_node = doc.getElementsByTagName(result_node_name);
    if (result_node.length !== 1) {
      throw new Error('Invalid XML from camera');
    }

    return result_node[0];
  };

  return Object.freeze({
    ruleList,
    ruleAdd,
    ruleRemove,
    checkEdgeReady,
  });
};